 /* Copyright (C) 2007, 2008 Anders Holst
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2, or
  * (at your option) any later version.
  * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  */

 /*
  * reconstructap - Version 1, 081010
  *
  * Compiles with the command
  *   g++ -O3 -o mreconstructap mreconstructap.cc
  */

#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LEN 24064

inline unsigned long long int byteswopl(unsigned long long int n)
{
  return (n>>56) | ((n>>40)&0xff00) | ((n>>24)&0xff0000) | ((n>>8)&0xff000000) | ((n&0xff000000)<<8) | ((n&0xff0000)<<24) | ((n&0xff00)<<40) | ((n&0xff)<<56);
}

int writebufinternal(int f, off64_t sz, off64_t tm)
{
  off64_t buf[2];
  buf[0] = (off64_t)byteswopl((unsigned long long int)sz);
  buf[1] = (off64_t)byteswopl((unsigned long long int)tm);
  if (write(f, buf, 16) != 16)
    return 1;
  else
    return 0;
}

int framepid(char* buf, int pos)
{
  return ((buf[pos+1] & 0x1f) << 8) + buf[pos+2];
}

off64_t framepts(char* buf, int pos)
{
  int tmp = (buf[pos+3] & 0x20 ? pos+buf[pos+4]+5 : pos+4);
  off64_t pts;
  if (buf[pos+1] & 0x40 &&
      buf[pos+3] & 0x10 &&
      buf[tmp]==0 && buf[tmp+1]==0 && buf[tmp+2]==1 &&
      buf[tmp+7] & 0x80) {
    pts  = ((unsigned long long)(buf[tmp+9]&0xE))  << 29;
    pts |= ((unsigned long long)(buf[tmp+10]&0xFF)) << 22;
    pts |= ((unsigned long long)(buf[tmp+11]&0xFE)) << 14;
    pts |= ((unsigned long long)(buf[tmp+12]&0xFF)) << 7;
    pts |= ((unsigned long long)(buf[tmp+13]&0xFE)) >> 1;
  } else
    pts = 0;
  return pts;
}

int framesearch(int fts, off64_t& retpos, off64_t& retpts)
{
  static char buf[LEN];
  static int ind;
  static off64_t pos = -1;
  static off64_t num;
  static int pid = -1;
  int st, count, tp=0;
  char* p;
  if (pos == -1) {
    num = read(fts, buf, LEN);
    ind = 0;
    pos = 0;
    st = 0;
  } else
    st = 188;
  count = 0;
  while (1) {
    p = buf+ind+st;
    ind = -1;
    for (; p < buf+num-5; p++)
      if (p[0]==0 && p[1]==0 && p[2]==1 && p[3]==0) {
        ind = ((p - buf)/188)*188;
        if (pid != -1 && pid != framepid(buf, ind)) {
          ind = -1;
          continue;
        }
        tp = (p[5]>>3)&7;
        if (pid == -1 && tp == 1)
          pid = framepid(buf, ind);
        break;
      }
    st = 0;
    if (ind == -1) {
      if (num == LEN) {
        pos += num;
        num = read(fts, buf, LEN);
        ind = 0;
      } else if (num) {
        ind = num;
        retpts = 0;
        retpos = pos + num;
        num = 0;
        return -1;
      } else {
        retpts = 0;
        retpos = 0;
        return -1;
      }
    } else {
      count++;
      st = 188;
      if (tp == 1) {
        retpts = framepts(buf, ind);
        retpos = pos + ind;
        return count; 
      }
    }
  }
}

int doit(int fao, int fts)
{
  off64_t pos;
  off64_t pts;
  while (framesearch(fts, pos, pts) >= 0)
    if (writebufinternal(fao, pos, pts))
      return 1;
  return 0;
}

char* makefilename(const char* base, const char* pre, const char* ext, const char* post)
{
  static char buf[256];
  int len1, len2, len3;
  len1 = strlen(base);
  len2 = (pre ? strlen(pre) : 0);
  len3 = (ext ? strlen(ext) : 0);
  strcpy(buf, base);
  if (ext && len1>=len3 && !strcmp(base+len1-len3,ext))
    len1 -= len3;
  if (pre)
    strcpy(buf+len1, pre);
  if (ext)
    strcpy(buf+len1+len2, ext);
  if (post)
    strcpy(buf+len1+len2+len3, post);
  return buf;
}

int main(int argc, char* argv[])
{
  int f_ts, f_apout;
  char* tmpname;
  char* inname = 0;

  if (argc != 2) {
    printf("Usage: mreconstructap ts_file\n");
    exit(1);
  }
  inname = argv[1];

  tmpname = makefilename(inname, 0, ".ts", 0);
  f_ts = open(tmpname, O_RDONLY | O_LARGEFILE);
  if (f_ts == -1) {
    printf("Failed to open input stream file \"%s\"\n", tmpname);
    exit(1);
  }
  tmpname = makefilename(inname, 0, ".ts", ".ap");
  f_apout = open(tmpname, O_WRONLY | O_CREAT | O_TRUNC, 0x1a4);
  if (f_apout == -1) {
    printf("Failed to open output ap file \"%s\"\n", tmpname);
    close(f_ts);
    exit(1);
  }

  doit(f_apout, f_ts);

  close(f_ts);
  close(f_apout);
}

