 /* Copyright (C) 2007, 2008 Anders Holst
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2, or
  * (at your option) any later version.
  * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  */

 /*
  * msplice - Version 1, 081010
  *
  * Compiles with the command
  *   g++ -O3 -o msplice msplice.cc
  */

#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


struct MovieDisc {
  MovieDisc(off64_t p1_ts, off64_t p2_ts, off64_t p1_ap, off64_t p2_ap, off64_t t) { begts = p1_ts, begap = p1_ap, endts = p2_ts, endap = p2_ap, tm = t, uncleancut = (begts != endts), next = 0; };
  off64_t begts, endts, begap, endap, tm;
  int uncleancut;
  MovieDisc* next;
};

struct MovieDescr {
  MovieDescr(char* name);
  ~MovieDescr();
  int exists();
  int open_in();
  int open_out();
  int close();
  int write(MovieDescr* out, int sec);
  char *name_ts, *name_ap, *alias;
  int f_ts, f_ap;
  off64_t size;
  int numsecs;
  int ap_offset;
  MovieDisc* discs;
  MovieDescr* next;
};

struct MovieSection {
  MovieSection(MovieDescr* md, int sec) { movie = md; section = sec; next = 0; };
  MovieDescr* movie;
  int section;
  MovieSection* next;
};

#define LEN 24064

static off64_t* buf0 = 0;
static off64_t* buf1 = 0;

inline int inbetween(long long int x, int lim1, int lim2)
{
  return (x>=lim1 && x<lim2);
}

char* lltimestr(long long int t)
{
  static char buf[15];
  int r = t%(90000*60);
  int m = t/(90000*60);
  sprintf(buf, "%d:%d:%.3f", m/60, m%60, r/90000.0);
  return buf;
}

char* lltimestr1(long long int t)
{
  static char buf[15];
  int r = t%(90000*60);
  int m = t/(90000*60);
  sprintf(buf, "%d:%d", m, (r+45000)/90000);
  return buf;
}

inline unsigned long long int byteswopl(unsigned long long int n)
{
  return (n>>56) | ((n>>40)&0xff00) | ((n>>24)&0xff0000) | ((n>>8)&0xff000000) | ((n&0xff000000)<<8) | ((n&0xff0000)<<24) | ((n&0xff00)<<40) | ((n&0xff)<<56);
}

int readbufinternal(int f, int offset)
{
  off64_t* buf;
  buf = buf0;
  buf0 = buf1;
  buf1 = buf;
  if (read(f, buf, 16) != 16)
    return 0;
  buf[0] = (off64_t)byteswopl((unsigned long long int)buf[0]) - offset;
  buf[1] = (off64_t)byteswopl((unsigned long long int)buf[1]);
  return 1;
}

void writebufinternal(int f, int offset)
{
  off64_t buf2[2];
  buf2[0] = (off64_t)byteswopl((unsigned long long int)buf1[0] + offset);
  buf2[1] = (off64_t)byteswopl((unsigned long long int)buf1[1]);
  write(f, buf2, 16);
}

int framepid(char* buf, int pos)
{
  return ((buf[pos+1] & 0x1f) << 8) + buf[pos+2];
}

off64_t framepts(char* buf, int pos)
{
  int tmp = (buf[pos+3] & 0x20 ? pos+buf[pos+4]+5 : pos+4);
  off64_t pts;
  if (buf[pos+1] & 0x40 &&
      buf[pos+3] & 0x10 &&
      buf[tmp]==0 && buf[tmp+1]==0 && buf[tmp+2]==1 &&
      buf[tmp+7] & 0x80) {
    pts  = ((unsigned long long)(buf[tmp+9]&0xE))  << 29;
    pts |= ((unsigned long long)(buf[tmp+10]&0xFF)) << 22;
    pts |= ((unsigned long long)(buf[tmp+11]&0xFE)) << 14;
    pts |= ((unsigned long long)(buf[tmp+12]&0xFF)) << 7;
    pts |= ((unsigned long long)(buf[tmp+13]&0xFE)) >> 1;
  } else
    pts = 0;
  return pts;
}

int framesearch_f(char* buf, int start, int stop, int pid, off64_t& pts, int& tp)
{
  char* p;
  int ind = -1;
  for (p = buf+start; p < buf+stop-5; p++)
    if (p[0]==0 && p[1]==0 && p[2]==1 && p[3]==0) {
      ind = ((p - buf)/188)*188;
      if (pid == -1 || framepid(buf, ind) == pid) {
        tp = (p[5]>>3)&7;
        pts = framepts(buf, ind);
        return ind;
      }
    }
  return -1;
}

off64_t framesearch_b(char* buf, int start, int stop, int pid, off64_t& pts, int& tp)
{
  char* p;
  int ind = -1;
  for (p = buf+stop-3; p >= buf+start+3; p--)
    if (p[0]==0 && p[-1]==1 && p[-2]==0 && p[-3]==0) {
      ind = ((p - buf)/188)*188;
      if (pid == -1 || framepid(buf, ind) == pid) {
        pts = framepts(buf, ind);
        tp = (p[2]>>3)&7;
        return ind;
      }
    }
  return -1;
}

off64_t framesearch_beg(int f, off64_t start, off64_t stop, int& uncl)
{
  int ind;
  char buf[LEN];
  off64_t pts1, pts2, pos, last, tmp;
  int tp, pid;
  pos = start - LEN + 188;
  tmp = LEN;
  if (pos < stop)
    tmp += pos-stop, pos = stop;
  lseek64(f, pos, SEEK_SET);
  if (read(f, buf, tmp) != tmp) return -1;
  ind = framesearch_b(buf, tmp > 376 ? tmp-376 : 0, tmp, -1, pts1, tp);
  if (ind == -1 || tp != 1)
    return -1; // ap felaktig
  start = pos + ind;
  pid = framepid(buf, ind);
  tmp = (tmp > 376 ? tmp-376 : 0);
  while (tmp) {
    ind = framesearch_b(buf, 0, tmp, pid, pts2, tp);
    if (ind != -1) {
      if (!pts1 || !pts2 || inbetween(pts1-pts2, 0, 180000)) {
        uncl = 0;
        return pos+ind;
      } else
        break;
    }
    pos -= LEN;
    tmp = LEN;
    if (pos < stop)
      tmp = LEN+pos-stop, pos = stop;
    lseek64(f, pos, SEEK_SET);
    if (read(f, buf, tmp) != tmp) return -1;
  }
  uncl = 1;
  pos = start + 188;
  lseek64(f, pos, SEEK_SET);
  ind = -1;
  tmp = 0;
  last = start;
  do {
    if (ind == -1) {
      pos += tmp;
      tmp = read(f, buf, LEN);
      ind = 0;
    } else
      ind += 188;
    ind = framesearch_f(buf, ind, tmp, pid, pts2, tp);
    if (ind != -1) {
      if (tp == 1)
        return last;
      last = pos + ind;
    }
  } while (tmp == LEN || ind != -1);
  return -1;
}

off64_t framesearch_end(int f, off64_t start, off64_t stop, int& uncl)
{
  int ind;
  char buf[LEN];
  off64_t pts1, pts2, pos, last, tmp;
  int tp, pid;
  pos = (start < 188 ? 0 : start - 188);
  lseek64(f, pos, SEEK_SET);
  tmp = read(f, buf, LEN);
  ind = framesearch_f(buf, 0, tmp < 376 ? tmp : 376, -1, pts1, tp);
  if (start == stop) {
    uncl = 1;
    if (ind == -1)
      if (tmp < 376)
        pid = -1;
      else
        return -1;
    else if (tp != 1)
      return -1;
    else
      pid = framepid(buf, ind);
    last = (ind != -1 ? pos+ind : start);
    pos = pos - LEN;
    tmp = LEN;
    if (pos < 0)
      tmp += pos, pos = 0;
    lseek64(f, pos, SEEK_SET);
    if (read(f, buf, tmp) != tmp) return -1;
    while (pos || tmp) {
      ind = framesearch_b(buf, 0, tmp, pid, pts2, tp);
      if (ind != -1) {
        if (tp==1)
          return last;
        else if (tp==2)
          last = pos+ind;
        tmp = ind;
        continue;
      }
      pos -= LEN;
      tmp = LEN;
      if (pos < 0)
        tmp += pos, pos = 0;
      lseek64(f, pos, SEEK_SET);
      if (read(f, buf, tmp) != tmp) return -1;
    }
  } else {
    if (ind == -1 || tp != 1)
      return -1; // ap felaktig
    pid = framepid(buf, ind);
    do {
      if (ind == -1) {
        pos += tmp;
        tmp = read(f, buf, LEN);
        ind = 0;
      } else
        ind += 188;
      ind = framesearch_f(buf, ind, tmp, pid, pts2, tp);
      if (pos+(ind==-1 ? tmp : ind) >= stop) {
        uncl = 0;
        return stop;
      } else if (ind != -1 && (tp != 3 || (pts1 && pts2 && !inbetween(pts1-pts2, 0, 180000)))) {
        uncl = 1;
        return pos+ind;
      }
    } while (tmp == LEN || ind != -1);
  }
  return -1;
}

char* makefilename(const char* base, const char* pre, const char* ext, const char* post)
{
  char* buf;
  int len1, len2, len3, len4;
  len1 = strlen(base);
  len2 = (pre ? strlen(pre) : 0);
  len3 = (ext ? strlen(ext) : 0);
  len4 = (post ? strlen(post) : 0);
  buf = new char[len1+len2+len3+len4+1];
  strcpy(buf, base);
  if (ext && len1>=len3 && !strcmp(base+len1-len3,ext))
    len1 -= len3;
  if (pre)
    strcpy(buf+len1, pre);
  if (ext)
    strcpy(buf+len1+len2, ext);
  if (post)
    strcpy(buf+len1+len2+len3, post);
  return buf;
}

MovieSection* parsefilearg(char* arg, MovieDescr*& movies, MovieSection*& last)
{
  // Check for alias, "=" and movie exists
  // Check sections, ":" and alias or movie exists  -> error if wrong sections
  // Check movie only, movie exists 
  // else error
  char *p, *tmp;
  MovieDescr* md;
  MovieSection *ms1, *ms2;
  int s1, s2;
  for (p=arg; *p && *p!='=' && *p!=':'; p++);
  if (*p=='=') {
    md =  new MovieDescr(p+1);
    if (md->exists() > 0) {
      if (md->open_in()) {
        //printf("Failed to open input file \"%s\"\n", p+1);
        exit(1);
      }
      *p = 0;
      md->alias = arg;
      md->next = movies;
      movies = md;
      return 0;
    } else
      delete md;
  }
  for (; *p; p++);
  for (p--; p>arg && ((*p>='0' && *p<='9') || *p==',' || *p=='-'); p--);
  if (*p == ':' && p != arg && *(p+1)) {
    *p = 0;
    for (md = movies; md && (!md->alias || strcmp(md->alias,arg)); md=md->next);
    if (!md) {
      tmp = makefilename(arg, 0, ".ts", 0);
      for (md = movies; md && strcmp(md->name_ts,tmp); md=md->next);
      delete [] tmp;
    }
    if (!md) {
      md =  new MovieDescr(arg);
      if (md->exists() > 0) {
        if (md->open_in()) {
          //printf("Failed to open input file \"%s\"\n", arg);
          exit(1);
        }
        md->next = movies;
        movies = md;
      } else {
        delete md;
        md = 0;
        *p = ':';
      }
    }
    if (md) {
      p++;
      ms1 = ms2 = 0;
      while (1) {
        if (*p>='1' && *p<='9') {
          s1 = atoi(p);
          for (; *p>='0' && *p<='9'; p++);
        } else if (*p == '-')
          s1 = 1;
        else {
          printf("Malformed section list for movie %s\n", arg);
          exit(1);
        }
        if (s1<=md->numsecs) {
          if (ms2)
            ms2 = ms2->next = new MovieSection(md, s1);
          else
            ms2 = ms1 = new MovieSection(md, s1);
        } else {
          printf("Bad section range for movie %s\n", arg);
          exit(1);
        }
        if (*p=='-') {
          p++;
          if (*p>='1' && *p<='9') {
            s2 = atoi(p);
            for (; *p>='0' && *p<='9'; p++);
          } else if (*p == ',')
            s2 = md->numsecs;
          else {
            printf("Malformed section list for movie %s\n", arg);
            exit(1);
          }
          if (s2>md->numsecs) {
            printf("Bad section range for movie %s\n", arg);
            exit(1);
          }
          else if (s2<=s1) {
            printf("Malformed section list for movie %s\n", arg);
            exit(1);
          }
          else {
            for (s1++; s1<=s2; s1++)
              ms2 = ms2->next = new MovieSection(md, s1);
          }
        }
        if (*p==',')
          p++;
        else if (*p==0)
          break;
        else {
          printf("Malformed section list for movie %s\n", arg);
          exit(1);
        }
      }
      last = ms2;
      return ms1;
    }
  }
  tmp = makefilename(arg, 0, ".ts", 0);
  for (md = movies; md && strcmp(md->name_ts,tmp); md=md->next);
  delete [] tmp;
  if (!md) {
    md =  new MovieDescr(arg);
    if (md->exists() > 0) {
      if (md->open_in()) {
        //printf("Failed to open input file \"%s\"\n", arg);
        exit(1);
      }
      md->next = movies;
      movies = md;
    } else {
      delete md; {
      printf("No file found for specification \"%s\"\n", arg);
      exit(1);
      }
    }
  }
  ms2 = ms1 = new MovieSection(md, 1);
  for (s1=2; s1<=md->numsecs; s1++)
    ms2 = ms2->next = new MovieSection(md, s1);
  last = ms2;
  return ms1;
}

char* readmeta(char* name, int& n)
{
  char* buf;
  char* tmpname = makefilename(name, 0, ".ts", ".meta");
  int f = open(tmpname, O_RDONLY);
  if (f == -1) {
    //printf("Failed to open input meta file \"%s\"\n", tmpname);
    delete [] tmpname;
    return 0;
  }
  n = lseek(f, 0, SEEK_END);
  lseek(f, 0, SEEK_SET);
  buf = new char[n];
  if (read(f, buf, n) != n) {
    printf("Failed to read input meta file \"%s\"\n", tmpname);
    delete [] tmpname;
    delete [] buf;
    close(f);
    return 0;
  }
  close(f);
  delete [] tmpname;
  return buf;
}

void writemeta(char* name, int n, const char* buf, const char* title, const char* suff, const char* descr)
{
  int i, j, k;
  char* tmpname = makefilename(name, 0, ".ts", ".meta");
  int f = open(tmpname, O_WRONLY | O_CREAT | O_TRUNC, 420);
  if (f == -1) {
    printf("Failed to open output meta file \"%s\"\n", tmpname);
    delete [] tmpname;
    return;
  }
  for (i=0; i<n; i++)
    if (buf[i] == 10)
      break;
  write(f, buf, i);
  if (i == n) {
    close(f);
    return;
  }
  for (j=i+1; j<n; j++)
    if (buf[j] == 10)
      break;
  if (title) {
    write(f, buf+i, 1);
    for (k=0; title[k] && title[k] != 10; k++);
    write(f, title, k);
  } else {
    write(f, buf+i, j-i);
    if (suff && j-i>1)
      write(f, suff, strlen(suff));
  }
  if (j == n) {
    close(f);
    return;
  }
  i = j;
  for (j=i+1; j<n; j++)
    if (buf[j] == 10)
      break;
  if (descr) {
    write(f, buf+i, 1);
    for (k=0; descr[k] && descr[k] != 10; k++);
    write(f, descr, k);
  } else {
    write(f, buf+i, j-i);
  }
  if (j < n)
    write(f, buf+j, n-j);
  close(f);
}

MovieDescr::MovieDescr(char* name)
{
  name_ts = makefilename(name, 0, ".ts", 0);
  name_ap = makefilename(name, 0, ".ts", ".ap");
  alias = 0;
  size = 0;
  numsecs = 0;
  ap_offset = 0;
  discs = 0;
}

MovieDescr::~MovieDescr()
{
  MovieDisc *d1, *d2;
  delete [] name_ts;
  delete [] name_ap;
  for (d1=discs; d1; d1=d2) {
    d2 = d1->next;
    delete d1;
  }
}

int MovieDescr::exists()
{
  struct stat64 s64;
  if (stat64(name_ts, &s64) < 0)
    return 0;
  else if (stat64(name_ap, &s64) < 0)
    return -1;
  else
    return 1;
}

int MovieDescr::open_in()
{
  int apos, ok, uncl;
  off64_t time_offset, fc1, fc2;
  MovieDisc* lastdisc;
  // ppna ts och ap
  // Sk efter diskont, lgg in i sections
  f_ts = open(name_ts, O_RDONLY | O_LARGEFILE);
  if (f_ts == -1) {
    printf("Failed to open input stream file \"%s\"\n", name_ts);
    return 1;
  }
  f_ap = open(name_ap, O_RDONLY);
  if (f_ap == -1) {
    printf("Failed to open input ap file \"%s\"\n", name_ap);
    ::close(f_ts);
    return 1;
  }
  numsecs = 0;
  discs = 0;
  if (!buf0) {
    buf0 = new off64_t[2];
    buf1 = new off64_t[2];
  }
  size = lseek64(f_ts, 0, SEEK_END);
  lseek(f_ap, -16, SEEK_END);
  ok = readbufinternal(f_ap, 0);
  lseek(f_ap, 0, SEEK_SET);
  ok &= readbufinternal(f_ap, 0);
  ap_offset = (buf0[0] - buf1[0] == size ? buf1[0] : 0);
  buf1[0] -= ap_offset;
  ok &= readbufinternal(f_ap, ap_offset);
  if (!ok) {
    printf("The corresponding \".ap\"-file is empty.\n");
    ::close(f_ts);
    ::close(f_ap);
    return 1;
  }
  time_offset = buf0[1];
  if (buf1[1] > buf0[1] && buf1[1] - buf0[1] < 450000)
    time_offset -= (buf1[1]-buf0[1])*buf0[0]/(buf1[0]-buf0[0]);
  else if (inbetween(buf1[1] + ((long long int)1)<<33 - buf0[1], 0, 450000))
    time_offset -= (buf1[1] + ((long long int)1)<<33 -buf0[1])*buf0[0]/(buf1[0]-buf0[0]);
  fc1 = framesearch_beg(f_ts, buf0[0], 0, uncl);
  if (fc1 == -1) {
    printf("The \".ap\"-file is corrupt. Try running \"mreconstructap\" first.\n");
    ::close(f_ts);
    ::close(f_ap);
    return 1;
  }
  apos = 0;
  lastdisc = discs = new MovieDisc(fc1, 0, (fc1>buf0[0] ? apos+16 : apos), apos, buf0[1] - time_offset);
  while (ok) {
    apos += 16;
    if (buf1[1] < buf0[1] || buf1[1] - buf0[1] > 450000) {
      if (inbetween(buf1[1] + ((long long int)1)<<33 - buf0[1], 0, 450000))
	time_offset -= ((long long int)1)<<33;
      else {
	time_offset += buf1[1] - buf0[1];
        fc1 = framesearch_beg(f_ts, buf1[0], buf0[0], uncl);
        fc2 = framesearch_end(f_ts, buf0[0], (fc1 < buf1[0] ? fc1 : buf1[0]), uncl);
        if (fc1 == -1 || fc2 == -1) {
          printf("The \".ap\"-file is corrupt. Try running \"mreconstructap\" first.\n");
          ::close(f_ts);
          ::close(f_ap);
          return 1;
        }
        lastdisc = lastdisc->next = new MovieDisc(fc1, fc2, (fc1>buf1[0] ? apos+16 : apos), (fc2<buf0[0] ? apos-16 : apos), buf1[1] - time_offset);
        numsecs++;
      }
    }
    ok = readbufinternal(f_ap, ap_offset);
  }
  apos += 16;
  fc2 = framesearch_end(f_ts, buf0[0], size, uncl);
  if (fc2 == -1) {
    printf("The \".ap\"-file is corrupt. Try running \"mreconstructap\" first.\n");
    ::close(f_ts);
    ::close(f_ap);
    return 1;
  }
  lastdisc->next = new MovieDisc(size, fc2, apos, (fc2<buf0[0] ? apos-16 : apos), buf0[1] - time_offset);
  numsecs++;
  return 0;
}

int MovieDescr::open_out()
{
  f_ts = open(name_ts, O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE, 420);
  if (f_ts == -1) {
    printf("Failed to open output stream file \"%s\"\n", name_ts);
    return 1;
  }
  f_ap = open(name_ap, O_WRONLY | O_CREAT | O_TRUNC, 420);
  if (f_ap == -1) {
    printf("Failed to open output ap file \"%s\"\n", name_ap);
    ::close(f_ts);
    return 1;
  }
  size = 0;
  return 0;
}

int MovieDescr::close()
{
  ::close(f_ts);
  ::close(f_ap);
  return 0;
}

int MovieDescr::write(MovieDescr* out, int sec)
{
  int j;
  off64_t i, n;
  MovieDisc *d1, *d2;
  char buf[LEN];
  if (sec < 1 || sec > numsecs)
    return 1;
  for (d1=discs, j=1; j<sec; d1=d1->next, j++);
  d2 = d1->next;
  i = d1->begap;
  n = d2->endap;
  lseek(f_ap, i, SEEK_SET);
  for (; i<n; i+=16) {
    if (!readbufinternal(f_ap, ap_offset)) return 1;
    writebufinternal(out->f_ap, out->size - d1->begts);
  }
  i = d1->begts;
  n = (d2->uncleancut ? d2->endts : d2->begts);
  lseek64(f_ts, i, SEEK_SET);
  for (; i+LEN<=n; i+=LEN) {
    if (read(f_ts, buf, LEN) != LEN) return 1;
    if (::write(out->f_ts, buf, LEN) != LEN) return 1;
    out->size += LEN;
  }
  if (i < n) {
    if (read(f_ts, buf, n-i) != n-i) return 1;
    if (::write(out->f_ts, buf, n-i) != n-i) return 1;
    out->size += n-i;
  }
  return 0;
}

int main(int argc, char* argv[])
{
// Kolla i argv vad som ska gras, -o/-s/-l/-r
  char* inname = 0;
  char* outname = 0;
  char* metabuf = 0;
  char* tmpname;
  char tbuf[10];
  int repair = 0, append = 0, split = 0, list = 0;
  MovieDescr *movies, *mout;
  MovieSection *secs, *ls, *tmps;
  MovieDisc *d;
  int i, innames = 0, bad = 0, nmeta;
  for (i=1; i<argc; i++) {
    if (!strcmp(argv[i], "-l"))
      list = 1;
    else if (!strcmp(argv[i], "-s"))
      split = 1;
    else if (!strcmp(argv[i], "-r"))
      repair = 1;
    else if (!strcmp(argv[i], "-o")) {
      if (i == argc-1) {
        bad = 1;
        break;
      }
      append = 1;
      outname = argv[++i];
    } else if (*argv[i] == '-' && (*(argv[i]+1) == 0 ||*(argv[i]+2) == 0)) {
      bad = 1;
      break;
    } else {
      innames++;
      if (!inname)
        inname = argv[i];
    }
  }
  if (list+split+repair+append != 1 || (!append && innames != 1))
    bad = 1;
  if (bad) {
    printf("Usage: msplice -o outfile moviefile1[:n1,n2,...] [moviefile2[:m1,m2,...]] ...]\n");
    printf("               : Append the given files, or given sections of the files.\n");
    printf("       msplice -l moviefile       : List sections in the file.\n");
    printf("       msplice -s moviefile       : Split sections into separate files.\n");
    printf("       msplice -r moviefile       : Repair files with unclean discontinuities.\n");
    exit(1);
  }
  if (append) {
    movies = 0;
    secs = 0;
    for (i=1; i<argc; i++) {
      if (!strcmp(argv[i], "-o")) {
        i++;
        continue;
      }
      if (!secs)
        secs = parsefilearg(argv[i], movies, ls);
      else {
        ls->next = parsefilearg(argv[i], movies, tmps);
        ls = tmps;
      }
    }
    mout = new MovieDescr(outname);
    if (mout->open_out()) {
      //printf("Failed to open output file %s\n", outname);
      exit(1);
    }
    if (secs)
      metabuf = readmeta(secs->movie->name_ts, nmeta);
    while (secs) {
      secs->movie->write(mout, secs->section);
      tmps = secs;
      secs = secs->next;
      delete tmps;
    }
    mout->close();
    delete mout;
    if (metabuf) {
      writemeta(outname, nmeta, metabuf, "", 0, "");
      delete [] metabuf;
    }
    while (movies) {
      movies->close();
      mout = movies;
      movies = movies->next;
      delete mout;
    }
  } else {
    movies = new MovieDescr(inname);
    if (!(movies->exists() > 0)) {
      printf("Input file \"%s\" not found\n", inname);
      exit(1);
    }
    if (movies->open_in()) {
      //printf("Failed to open input file %s\n", inname);
      exit(1);
    }
    if (list) {
      printf("%d section%s:\n", movies->numsecs, (movies->numsecs==1 ? "" : "s"));
      for (i=1, d=movies->discs; d && d->next; d=d->next, i++) {
        if (d->uncleancut)
          printf("[Unclean cut at %s]\n", lltimestr(d->tm));
        printf("%d: %s (%s - ", i, lltimestr1(d->next->tm - d->tm), lltimestr(d->tm));
        printf("%s)\n", lltimestr(d->next->tm));
      }
      if (d && d->uncleancut)
        printf("[Unclean cut at %s]\n", lltimestr(d->tm));
    } else if (split) {
      metabuf = readmeta(inname, nmeta);
      for (i=1; i<=movies->numsecs; i++) {
        sprintf(tbuf, " sec%03d", i);
        outname = makefilename(inname, tbuf, ".ts", 0);
        mout = new MovieDescr(outname);
        if (mout->open_out()) {
          //printf("Failed to open output file %s\n", outname);
          exit(1);
        }
        movies->write(mout, i);
        mout->close();
        delete mout;
        if (metabuf)
          writemeta(outname, nmeta, metabuf, 0, tbuf, 0);
        delete [] outname;
      }
      if (metabuf)
        delete [] metabuf;
    } else if (repair) {
      if (movies->ap_offset)
        bad = 1;
      else {
        bad = 0;
        for (d=movies->discs; d; d=d->next)
          if (d->uncleancut)
            bad = 1;
      }
      if (!bad)
        printf("No need to repair \"%s\"\n", inname);
      else {
        outname = makefilename(inname, "_", ".ts", 0);
        mout = new MovieDescr(outname);
        if (mout->open_out()) {
          //printf("Failed to open output file %s\n", outname);
          exit(1);
        }
        delete [] outname;
        for (i=1; i<=movies->numsecs; i++)
          movies->write(mout, i);
        mout->close();
        tmpname = makefilename(inname, 0, ".ts", 0);
        rename(mout->name_ts, tmpname);
        delete [] tmpname; 
        tmpname = makefilename(inname, 0, ".ts", ".ap");
        rename(mout->name_ap, tmpname);
        delete [] tmpname; 
        delete mout;
      }
    }
    delete movies;
  }

// Om -o, parsa filnamn, annars bara ett namn
// Konstruera MovieDescr, dvs hitta sectionerna
// ppna ev. utfil och skriv
// Fixa meta-fil etc
}
